/*
 * dir.c
 */

#define EXPORT_SYMTAB


#define __NO_VERSION__
#include <linux/module.h>
#include <asm/uaccess.h>
#include <linux/sched.h>
#include <linux/stat.h>
#include <linux/string.h>
#include <linux/locks.h>
#include <linux/quotaops.h>
#include <linux/list.h>
#include <linux/file.h>
#include <asm/bitops.h>
#include <asm/byteorder.h>

#include <linux/filter.h>
#include <linux/snapfs.h>
#include <linux/snapsupport.h>

static ino_t get_parent_ino(struct inode * inode)
{
	ino_t ino = 0;
	struct dentry * dentry;

	if (list_empty(&inode->i_dentry)) {
       		printk("snapfs ERROR: no dentry for ino %lu\n", inode->i_ino);
                return 0;
        }

       	dentry = dget(list_entry(inode->i_dentry.next, struct dentry, d_alias));

        if(dentry->d_parent->d_inode)
		ino = dentry->d_parent->d_inode->i_ino;

	dput(dentry);
	return ino;

}

/* XXX check the return values */
static struct dentry *currentfs_lookup(struct inode * dir,struct dentry *dentry)
{
	struct snap_cache *cache;
	struct dentry *rc;
	struct inode_operations *iops;

	ENTRY;

	cache = snap_find_cache(dir->i_dev);
	if ( !cache ) { 
		EXIT;
		return ERR_PTR(-EINVAL);
	}

	if ( dentry->d_name.len == strlen(".snap") &&
	     (memcmp(dentry->d_name.name, ".snap", strlen(".snap")) == 0) ) {
		struct inode *snap;
		ino_t ino = 0xF0000000 | dir->i_ino;
		snap = iget(dir->i_sb, ino);
		CDEBUG(D_INODE, ".snap inode ino %ld, mode %o\n", snap->i_ino, snap->i_mode);
		d_add(dentry, snap);
		EXIT;
		return NULL;
	}

	iops = filter_c2cdiops(cache->cache_filter); 
	if (!iops || !iops->lookup) {
		EXIT;
		return ERR_PTR(-EINVAL);
	}
	rc = iops->lookup(dir, dentry);

#if 0
	/* XXX: PJB these need to be set up again. See dcache.c */
	printk("set up dentry ops\n");
	CDEBUG(D_CACHE, "\n");
        filter_setup_dentry_ops(cache->cache_filter,
                                dentry->d_op, &currentfs_dentry_ops);
        dentry->d_op = filter_c2udops(cache->cache_filter);
        CDEBUG(D_CACHE, "\n");
#endif

	EXIT;
	return NULL;
}

static int currentfs_create(struct inode *dir, struct dentry *dentry, int mode)
{
	struct snap_cache *cache;
	int rc;
	struct inode_operations *iops;
	void *handle = NULL;

	ENTRY;

#if 0
	/* XXX PJB: place this everywhere */ 
	if (currentfs_is_under_dotsnap(dentry)) {
		EXIT;
		return -EROFS;
	}
#endif

	cache = snap_find_cache(dir->i_dev);
	if ( !cache ) { 
		EXIT;
		return -EINVAL;
	}

	if ( snap_needs_cow(dir) != -1 ) {
		handle = snap_trans_start(cache, dir, SNAP_OP_CREATE);

		printk("snap_needs_cow for ino %lu \n",dir->i_ino);
		snap_debug_device_fail(dir->i_dev, SNAP_OP_CREATE, 1);
		snap_do_cow(dir, get_parent_ino(dir), 0);
	}

	iops = filter_c2cdiops(cache->cache_filter); 
	if (!iops ||
	    !iops->create) {
		rc = -EINVAL;
		goto exit;
	}
	snap_debug_device_fail(dir->i_dev, SNAP_OP_CREATE, 2);
	rc = iops->create(dir, dentry, mode);

	/* XXX now set the correct snap_{file,dir,sym}_iops */
	if ( S_ISDIR(dentry->d_inode->i_mode) )
                dentry->d_inode->i_op = filter_c2udiops(cache->cache_filter);
        else if ( S_ISREG(dentry->d_inode->i_mode) ) {
                if ( !filter_c2cfiops(cache->cache_filter) ) {
                        filter_setup_file_ops(cache->cache_filter,
                                dentry->d_inode->i_op, &currentfs_file_iops);
                }
                dentry->d_inode->i_op = filter_c2ufiops(cache->cache_filter);
        }
	printk("inode %lu, i_op %p\n", dentry->d_inode->i_ino, dentry->d_inode->i_op);

	snap_debug_device_fail(dir->i_dev, SNAP_OP_CREATE, 3);
	
exit:
	if( handle )
		snap_trans_commit(cache, handle);
	EXIT;
	return rc;
}

static int currentfs_mkdir(struct inode *dir, struct dentry *dentry, int mode)
{
	struct snap_cache *cache;
	int rc;
	struct inode_operations *iops;
	void *handle = NULL;

	ENTRY;

	cache = snap_find_cache(dir->i_dev);
	if ( !cache ) { 
		EXIT;
		return -EINVAL;
	}

	if ( snap_needs_cow(dir) != -1 ) {
		handle = snap_trans_start(cache, dir, SNAP_OP_MKDIR);

		printk("snap_needs_cow for ino %lu \n",dir->i_ino);
		snap_debug_device_fail(dir->i_dev, SNAP_OP_MKDIR, 1);
		snap_do_cow(dir, get_parent_ino(dir), 0);
	}

	iops = filter_c2cdiops(cache->cache_filter); 
	if (!iops ||
	    !iops->mkdir) {
		rc = -EINVAL;
		goto exit;
	}
	snap_debug_device_fail(dir->i_dev, SNAP_OP_MKDIR, 2);
	rc = iops->mkdir(dir, dentry, mode);

	/* XXX now set the correct snap_{file,dir,sym}_iops */
        dentry->d_inode->i_op = filter_c2udiops(cache->cache_filter);
	printk("inode %lu, i_op %p\n", dentry->d_inode->i_ino, dentry->d_inode->i_op);

	snap_debug_device_fail(dir->i_dev, SNAP_OP_MKDIR, 3);
	
exit:
	if( handle )
		snap_trans_commit(cache, handle);
	EXIT;
	return rc;
}

static int currentfs_link (struct dentry * old_dentry, struct inode * dir, 
			struct dentry *dentry)
{
	struct snap_cache *cache;
	int rc;
	struct inode_operations *iops;
	void *handle = NULL;

	ENTRY;

	cache = snap_find_cache(dir->i_dev);
	if ( !cache ) { 
		EXIT;
		return -EINVAL;
	}

	if ( snap_needs_cow(dir) != -1 ) {
		handle = snap_trans_start(cache, dir, SNAP_OP_LINK);

		printk("snap_needs_cow for ino %lu \n",dir->i_ino);
		snap_debug_device_fail(dir->i_dev, SNAP_OP_LINK, 1);
		snap_do_cow(dir, get_parent_ino(dir), 0);
	}

	iops = filter_c2cdiops(cache->cache_filter); 
	if (!iops ||
	    !iops->link) {
		rc = -EINVAL;
		goto exit;
	}
	snap_debug_device_fail(dir->i_dev, SNAP_OP_LINK, 2);
	rc = iops->link(old_dentry,dir, dentry);
	snap_debug_device_fail(dir->i_dev, SNAP_OP_LINK, 3);
	
exit:
	if( handle )
		snap_trans_commit(cache, handle);
	EXIT;
	return rc;
}

static int currentfs_symlink(struct inode *dir, struct dentry *dentry, 
			const char * symname)
{
	struct snap_cache *cache;
	int rc;
	struct inode_operations *iops;
	void *handle = NULL;

	ENTRY;

	cache = snap_find_cache(dir->i_dev);
	if ( !cache ) { 
		EXIT;
		return -EINVAL;
	}

	if ( snap_needs_cow(dir) != -1 ) {
		handle = snap_trans_start(cache, dir, SNAP_OP_SYMLINK);

		printk("snap_needs_cow for ino %lu \n",dir->i_ino);
		snap_debug_device_fail(dir->i_dev, SNAP_OP_SYMLINK, 1);
		snap_do_cow(dir, get_parent_ino(dir), 0);
	}

	iops = filter_c2cdiops(cache->cache_filter); 
	if (!iops ||
	    !iops->symlink) {
		rc = -EINVAL;
		goto exit;
	}
	snap_debug_device_fail(dir->i_dev, SNAP_OP_SYMLINK, 2);
	rc = iops->symlink(dir, dentry, symname);
	snap_debug_device_fail(dir->i_dev, SNAP_OP_SYMLINK, 3);
	
exit:
	if( handle )
		snap_trans_commit(cache, handle);
	EXIT;
	return rc;
}

static int currentfs_mknod(struct inode *dir, struct dentry *dentry, int mode, 
			int rdev)
{
	struct snap_cache *cache;
	int rc;
	struct inode_operations *iops;
	void *handle = NULL;

	ENTRY;

	cache = snap_find_cache(dir->i_dev);
	if ( !cache ) { 
		EXIT;
		return -EINVAL;
	}

	if ( snap_needs_cow(dir) != -1 ) {
		handle = snap_trans_start(cache, dir, SNAP_OP_MKNOD);

		printk("snap_needs_cow for ino %lu \n",dir->i_ino);
		snap_debug_device_fail(dir->i_dev, SNAP_OP_MKNOD, 1);
		snap_do_cow(dir, get_parent_ino(dir), 0);
	}

	iops = filter_c2cdiops(cache->cache_filter); 
	if (!iops ||
	    !iops->mknod) {
		rc = -EINVAL;
		goto exit;
	}
	snap_debug_device_fail(dir->i_dev, SNAP_OP_MKNOD, 2);
	rc = iops->mknod(dir, dentry, mode, rdev);
	snap_debug_device_fail(dir->i_dev, SNAP_OP_MKNOD, 3);
	
	/* XXX do we need to set the correct snap_{*}_iops */

exit:
	if( handle )
		snap_trans_commit(cache, handle);
	EXIT;
	return rc;
}

static int currentfs_rmdir(struct inode *dir, struct dentry *dentry)
{
	struct snap_cache *cache;
	int rc;
	struct inode_operations *iops;
	struct inode *inode = NULL;
//	time_t i_ctime = 0;
	nlink_t	i_nlink = 0;
	off_t	i_size = 0;
	ino_t ino = 0;
	int keep_inode = 0;
//	struct dentry_operations *save_dop = NULL;
	void *handle = NULL;

	ENTRY;

	cache = snap_find_cache(dir->i_dev);
	if ( !cache ) { 
		EXIT;
		return -EINVAL;
	}

	if ( snap_needs_cow(dir) != -1 ) {
		handle = snap_trans_start(cache, dir, SNAP_OP_RMDIR);

		printk("snap_needs_cow for ino %lu \n",dir->i_ino);
		snap_debug_device_fail(dir->i_dev, SNAP_OP_RMDIR, 1);
		snap_do_cow(dir, get_parent_ino(dir), 0);
	}

	iops = filter_c2cdiops(cache->cache_filter); 
	if (!iops ||
	    !iops->rmdir) {
		rc = -EINVAL;
		goto exit;
	}

	/* XXX : there are two cases that we can't remove this inode from disk. 
		1. the inode needs to be cowed. 
		2. the inode is a redirector.
		then we must keep this inode(dir) so that the inode 
		will not be deleted after rmdir, will only remove dentry 
	*/

	if( snap_needs_cow(dentry->d_inode) != -1) {
		snap_debug_device_fail(dir->i_dev, SNAP_OP_RMDIR, 2);
		snap_do_cow (dentry->d_inode, dir->i_ino, 
				SNAP_DEL_PRI_WITHOUT_IND);
		keep_inode = 1;
	}
	else if( snap_is_redirector(dentry->d_inode) ) {
		snap_debug_device_fail(dir->i_dev, SNAP_OP_RMDIR, 3);
		snap_do_cow(dentry->d_inode, dir->i_ino, SNAP_DEL_PRI_WITH_IND);
		keep_inode = 1;
	}	
#if 0
	if ( keep_inode ) {	
       		printk("set up dentry ops, before %p\n",dentry->d_op);
		save_dop = dentry->d_op;

     		filter_setup_dentry_ops(cache->cache_filter,
                                dentry->d_op, &currentfs_dentry_ops);
        	dentry->d_op = filter_c2udops(cache->cache_filter);

       		printk("set up dentry ops, after %p\n",dentry->d_op);

	}

#endif

	if( keep_inode && dentry->d_inode ) {
		ino = dentry->d_inode->i_ino;
	//	i_ctime = dentry->d_inode->i_ctime;
		i_nlink = dentry->d_inode->i_nlink;
		i_size = dentry->d_inode->i_size;
	
}

	snap_debug_device_fail(dir->i_dev, SNAP_OP_RMDIR, 4);
	rc = iops->rmdir(dir, dentry);
	snap_debug_device_fail(dir->i_dev, SNAP_OP_RMDIR, 5);

	/* XXX : check this */
#if 0
	if ( keep_inode ) {
		dentry->d_op = save_dop;
		printk("restore dentry ops, now at %p\n",dentry->d_op);
	}

#endif

	if( keep_inode && ino) {
		inode = iget ( dir->i_sb, ino);
		if( inode) {
//			inode->i_ctime = i_ctime;
			inode->i_nlink = i_nlink;
			inode->i_size = i_size;
			mark_inode_dirty(inode);
			iput( inode);
			snap_debug_device_fail(dir->i_dev, SNAP_OP_RMDIR, 6);
		}
	}

exit:
	if( handle )
		snap_trans_commit(cache, handle);
	EXIT;
	return rc;
}

static int currentfs_unlink(struct inode *dir, struct dentry *dentry)
{
	struct inode *inode = dentry->d_inode;
	struct snap_cache *cache;
	int rc;
	struct inode_operations *iops;
	void *handle = NULL;

	ENTRY;

	cache = snap_find_cache(dir->i_dev);
	if ( !cache ) { 
		EXIT;
		return -EINVAL;
	}

	if ( snap_needs_cow(dir) != -1 ) {
		handle = snap_trans_start(cache, dir, SNAP_OP_UNLINK);

		printk("snap_needs_cow for ino %lu \n",dir->i_ino);
		snap_debug_device_fail(dir->i_dev, SNAP_OP_UNLINK, 1);
		snap_do_cow(dir, get_parent_ino(dir), 0);
	}

	iops = filter_c2cdiops(cache->cache_filter); 
	if (!iops ||
	    !iops->unlink) {
		rc = -EINVAL;
		goto exit;
	}

	/* XXX : if nlink for this inode is 1, there are two cases that we 
		can't remove this inode from disk. 
		1. the inode needs to be cowed. 
		2. the inode is a redirector.
		then we increament dentry->d_inode->i_nlink so that the inode 
		will not be deleted after unlink, will only remove dentry 
	*/

	if( inode->i_nlink == 1 ) {
		if( snap_needs_cow (inode) != -1) {
			/* call snap_do_cow with DEL_WITHOUT_IND option */
			snap_debug_device_fail(dir->i_dev, SNAP_OP_UNLINK, 2);
			snap_do_cow(inode, dir->i_ino,SNAP_DEL_PRI_WITHOUT_IND);
			inode->i_nlink++;
		}
		else if( snap_is_redirector (inode) ) {
			/* call snap_do_cow with DEL_WITH_IND option 
			 * just free the blocks of inode, not really delete it
			 */
			snap_debug_device_fail(dir->i_dev, SNAP_OP_UNLINK, 3);
			snap_do_cow (inode, dir->i_ino, SNAP_DEL_PRI_WITH_IND);
			inode->i_nlink++;
		}	
	}

	snap_debug_device_fail(dir->i_dev, SNAP_OP_UNLINK, 4);
	rc = iops->unlink(dir, dentry);
	snap_debug_device_fail(dir->i_dev, SNAP_OP_UNLINK, 5);

exit:
	if( handle )
		snap_trans_commit(cache, handle);
	EXIT;
	return rc;
}

static int currentfs_rename (struct inode * old_dir, struct dentry *old_dentry,
                      	struct inode * new_dir, struct dentry *new_dentry)
{
	struct snap_cache *cache;
	int rc;
	struct inode_operations *iops;
	void *handle = NULL;

	ENTRY;

	cache = snap_find_cache(old_dir->i_dev);
	if ( !cache ) { 
		EXIT;
		return -EINVAL;
	}

	handle = snap_trans_start(cache, old_dir, SNAP_OP_RENAME);

	if ( snap_needs_cow(old_dir) != -1 ) {
		printk("rename: needs_cow for old_dir %lu\n",old_dir->i_ino);
		snap_debug_device_fail(old_dir->i_dev, SNAP_OP_RENAME, 1);
		snap_do_cow(old_dir, get_parent_ino(old_dir), 0);
	}

	/* if they are not the same dir, and new_dentry is NULL, it's the
	case when you move a file to a dir, then cow the new_dir */
	if(( old_dir != new_dir) && ( !new_dentry->d_inode )) {
		printk("rename:new_dentry is NULL\n");
		if( snap_needs_cow(new_dir) !=-1 ){
			printk("rename:snap_needs_cow for new_dir %lu\n",
				new_dir->i_ino);
			snap_debug_device_fail(old_dir->i_dev, SNAP_OP_RENAME, 2);
			snap_do_cow(new_dir, get_parent_ino(new_dir),0);	
		}
	}

	if( ( old_dir != new_dir) && ( new_dentry->d_inode )) {
		if(snap_needs_cow(new_dentry->d_inode) !=-1 ){
			printk("rename:needs_cow for new_entry ,ino %lu\n",
				new_dentry->d_inode->i_ino);
			snap_debug_device_fail(old_dir->i_dev, SNAP_OP_RENAME, 3);
			snap_do_cow (new_dentry->d_inode, 
				new_dentry->d_parent->d_inode->i_ino, 0);	
		}
	}

	if( snap_needs_cow (old_dentry->d_inode) != -1) {
		snap_debug_device_fail(old_dir->i_dev, SNAP_OP_RENAME, 4);
		snap_do_cow(old_dentry->d_inode, old_dir->i_ino,0);
	}

	iops = filter_c2cdiops(cache->cache_filter); 
	if (!iops ||
	    !iops->rename) {
		rc = -EINVAL;
		goto exit;
	}

	snap_debug_device_fail(old_dir->i_dev, SNAP_OP_RENAME, 5);
	rc = iops->rename(old_dir, old_dentry, new_dir, new_dentry);
	snap_debug_device_fail(old_dir->i_dev, SNAP_OP_RENAME, 6);

exit:
	if( handle )
		snap_trans_commit(cache, handle);
	EXIT;
	return rc;
}

struct file_operations currentfs_dir_fops;
struct inode_operations currentfs_dir_iops = { 
	default_file_ops: &currentfs_dir_fops,
	create: currentfs_create,
	mkdir: currentfs_mkdir,
	link: currentfs_link,
	symlink: currentfs_symlink,
	mknod: currentfs_mknod,
	rmdir: currentfs_rmdir,
	unlink: currentfs_unlink,
	rename: currentfs_rename,
	lookup:	currentfs_lookup
};
