#ifndef __LINUX_SNAPFS_H
#define __LINUX_SNAPFS_H

#include <linux/list.h>
#include <linux/fs.h>
#include <linux/time.h>

#include <linux/filter.h>
#include <linux/snap.h>

/* sysctl.c */
extern int init_snapfs_proc_sys(void);
extern void cleanup_spapfs_proc_sys(void);

/* snap cache information: this morally equals the superblock of a
 snap_current_fs.  However, that superblock is the one of the "cache"
 device holding the inodes, hence we store this info in the hash of
 mountpoints hanging of our control device. 
*/
struct snap_cache {
	struct list_head cache_chain;

	kdev_t cache_dev;
	struct super_block *cache_sb; /* the _real_ device */

	struct list_head cache_clone_list;
	int cache_snap_tableno;

	struct filter_fs *cache_filter;

	char cache_type;
};

/* this is the snap_clone_info for the sb of snap_clone_fs */
struct snap_clone_info {
	struct snap_cache *clone_cache;
	struct list_head clone_list_entry;
	int clone_index;
};

#define SNAP_OP_NOOP		0
#define SNAP_OP_CREATE		1
#define SNAP_OP_MKDIR		2
#define SNAP_OP_UNLINK		3
#define SNAP_OP_RMDIR		4
#define SNAP_OP_CLOSE		5
#define SNAP_OP_SYMLINK		6
#define SNAP_OP_RENAME		7
#define SNAP_OP_SETATTR		8
#define SNAP_OP_LINK		9
#define SNAP_OP_OPEN		10
#define SNAP_OP_MKNOD		11
#define SNAP_OP_WRITE		12
#define SNAP_OP_RELEASE		13

struct journal_ops {
	void *(*trans_start)(struct inode *, int op);
	void (*trans_commit)(void *handle);
};

/* we have just a single snapshot control device
   it contains a list of all the snap_current info's
*/
#define SNAP_PSDEV_MINOR 240
#define SNAP_PSDEV_MAJOR 10
struct snap_control_device {
	struct list_head snap_dev_list;
};

/* ioctls for manipulating snapshots 40 - 60 */
#define IOC_SNAP_SETTABLE		_IOWR('f', 40, long)
#define IOC_SNAP_PRINTTABLE		_IOWR('f', 41, long)
#define IOC_SNAP_GETINDEXFROMNAME	_IOWR('f', 42, long)
#define IOC_SNAP_GET_NEXT_INO		_IOWR('f', 43, long)
#define IOC_SNAP_GET_INO_INFO		_IOWR('f', 44, long)

#define IOC_SNAP_DELETE			_IOWR('f', 45, long)
#define IOC_SNAP_RESTORE		_IOWR('f', 46, long)
#define IOC_SNAP_DEBUG			_IOWR('f', 47, long)
#define IOC_SNAP_DEVFAIL		_IOWR('f', 48, long)


/* What we use to point to IDs in the obdmd data for snapshots.  If we use
 * obd_id (8 bytes) instead of ino_t (4 bytes), we halve the number of
 * available snapshot slots (14 in 56 bytes vs. 7 in 56 bytes until we
 * increase the size of OBD_OBDMDSZ).
 */
typedef ino_t	snap_id;

/* maximum number of snapshot tables we maintain in the kernel */
#define SNAP_MAX_TABLES 32
#define SNAP_MAX_NAMELEN	64

/* maximum number of snapshots per device 
   must fit in "o_obdmd" area of struct obdo */
//#define OBD_OBDMDSZ  54
//#define SNAP_MAX ((OBD_OBDMDSZ - sizeof(uint32_t))/sizeof(snap_id))

#define SNAP_MAX	50

/* if time is 0 this designates the "current" snapshot, i.e.
   the head of the tree 
*/
struct snap {
	time_t time;
	int index;
	char name[SNAP_MAX_NAMELEN];
};

/* snap ioctl data for attach: current always in first slot of this array */
struct snap_obd_data {
	int 	     snap_dev;	/* which device contains the data */
	unsigned int snap_index;/* which snapshot is ours */
	unsigned int snap_table;/* which table do we use */
};

/* snap ioctl data for table fiddling */
struct snap_table_data {
	int 		tblcmd_no;	/* which table */
	unsigned int 	tblcmd_count;	/* how many snaps */
	struct snap 	tblcmd_snaps[SNAP_MAX];	/* sorted times! */
};


struct snap_table {
	spinlock_t          tbl_lock;
	unsigned int tbl_count; /* how many snapshots exist in this table*/
	int tbl_used;  /* bitmap of snaps in use by a device */
	time_t tbl_times[SNAP_MAX];
	int tbl_index[SNAP_MAX];
	char tbl_name[SNAP_MAX][SNAP_MAX_NAMELEN];
};

struct snap_iterdata {
//	struct obd_conn *conn;
//	struct obd_conn *ch_conn;
	kdev_t dev;	/* snap current device number */ 
	int index;
	int tableno;
//	int previndex;
//	int currentindex;
//	int prevslot;
//	time_t prevtime;
	time_t time;
};

struct snap_ioc_data {
	kdev_t dev;
	char name[SNAP_MAX_NAMELEN];
};

struct snap_ino_list_data{
        kdev_t dev;
        ino_t ino;
};

/* cache.c */
inline void snap_free_cache(struct snap_cache *cache);
struct snap_cache *snap_find_cache(kdev_t dev);

/* snaptable.c */
extern struct snap_table snap_tables[SNAP_MAX_TABLES];
void snap_last(struct snap_cache *info, struct snap *snap);
int snap_index2slot(struct snap_table *snap_table, int snap_index);
int snap_needs_cow(struct inode *);

/* snap.c */
int snap_is_redirector(struct inode *inode);
struct inode *snap_redirect(struct inode *inode, struct super_block *clone_sb);
int snap_do_cow(struct inode *inode, ino_t parent_ino, int del);

int snap_iterate(struct super_block *sb,
                int (*repeat)(struct inode *inode, void *priv),
                struct inode **start, void *priv, int flag);

struct inode *snap_get_indirect(struct inode *pri, int *table, int slot);
int snap_destroy_indirect(struct inode *pri, int index, struct inode *next_ind);
int snap_restore_indirect(struct inode *pri, int index );
int snap_migrate_data(struct inode *dst, struct inode *src);
int snap_set_indirect(struct inode *pri, ino_t ind_ino, 
			int index, ino_t parent_ino);

/* inode.c */
extern struct super_operations currentfs_super_ops;

/* dir.c */
extern struct inode_operations currentfs_dir_iops;
extern struct file_operations currentfs_dir_fops;

/* file.c */
extern struct inode_operations currentfs_file_iops;
extern struct file_operations currentfs_file_fops;

/* symlink.c */
extern struct inode_operations currentfs_sym_iops;
extern struct file_operations currentfs_sym_fops;

//#if 0
/* dcache.c */
extern struct dentry_operations currentfs_dentry_ops;
//#endif

#endif /* __LINUX_SNAPFS_H */
