#ifndef _SNAP_SUPPORT
#define _SNAP_SUPPORT
#include <linux/malloc.h>
#include <linux/vmalloc.h>
#include <linux/blk.h>

#define MIN(a,b) (((a)<(b)) ? (a): (b))
#define MAX(a,b) (((a)>(b)) ? (a): (b))

static inline void snapfs_cpy_attrs(struct inode *dst, struct inode *src)
{
	dst->i_mtime = src->i_mtime;
	dst->i_ctime = src->i_ctime;
	dst->i_atime = src->i_atime;
	dst->i_size = src->i_size;
	dst->i_blksize = src->i_blksize;
	dst->i_blocks = src->i_blocks;
	dst->i_generation = src->i_generation;
	dst->i_uid = src->i_uid;
	dst->i_gid = src->i_gid;
	dst->i_mode = src->i_mode;
}


/*
 * Debug code
 */
/* global variables */
extern int snap_debug_level;
extern int snap_print_entry;
extern int snap_inodes;
extern int snap_pages;
extern long snap_kmemory;
extern long snap_vmemory;

#define SNAP_DEBUG
#ifdef SNAP_DEBUG
#define CMD(cmd) (( cmd == READ ) ? "read" : "write")


#define D_SUPER     1   /* print results returned by Venus */
#define D_INODE     2   /* print entry and exit into procedure */
#define D_FILE      4
#define D_CACHE     8   /* cache debugging */
#define D_MALLOC    16  /* print malloc, de-alloc information */
#define D_SNAP      32
#define D_UPCALL    64  /* up and downcall debugging */
#define D_PSDEV    128
#define D_INFO     256
#define D_SPECIAL  512
#define D_IOCTL   1024
#define D_DOWNCALL 2048

#define CDEBUG(mask, format, a...)                                      \
	do {                                                            \
	if (snap_debug_level & mask) {                                   \
		printk("(%s:%d):", __FUNCTION__, __LINE__);             \
		printk(format, ## a); }                                 \
	} while (0)

#define ENTRY if (snap_print_entry)					\
	printk(KERN_INFO "Process %d entered %s\n", current->pid, __FUNCTION__)

#define EXIT if (snap_print_entry)					\
	printk(KERN_INFO "Process %d leaving %s [%d]\n", current->pid,	\
	       __FUNCTION__, __LINE__)

/* Inode common information printed out (used by snapfs and ext2snap inodes) */
#define ICDEBUG(inode) {						\
	CDEBUG(D_INFO,							\
	       "ino %ld, atm %ld, mtm %ld, ctm %ld, size %Ld, blocks %ld\n",\
	       inode->i_ino, inode->i_atime, inode->i_mtime, inode->i_ctime,\
	       inode->i_size, inode->i_blocks);				\
	CDEBUG(D_INFO, "mode %o, uid %d, gid %d, nlnk %d, count %d\n",	\
	       inode->i_mode, inode->i_uid, inode->i_gid, inode->i_nlink,\
	       inode->i_count);						\
}

/* Ext2 inode information */
#define EXDEBUG(inode) {						\
	ICDEBUG(inode);							\
	CDEBUG(D_INFO, "ext2 blocks: %d %d %d %d %d %d %d %d\n",	\
	       inode->u.ext2_i.i_data[0], inode->u.ext2_i.i_data[1],	\
	       inode->u.ext2_i.i_data[2], inode->u.ext2_i.i_data[3],	\
	       inode->u.ext2_i.i_data[4], inode->u.ext2_i.i_data[5],	\
	       inode->u.ext2_i.i_data[6], inode->u.ext2_i.i_data[7]);	\
}

/* SNAPFS inode information */
#define OIDEBUG(inode) {						\
	ICDEBUG(inode);							\
	CDEBUG(D_INFO,"oinfo: flags 0x%08x\n", snapfs_i2info(inode)->oi_flags);\
	/* snapfs_print_plist(inode); */					\
}


#else /* EXT2_SNAP_DEBUG */

#define CDEBUG(mask, format, a...) {}
#define ENTRY {}
#define EXIT {}
#define EXDEBUG(inode) {}
#define OIDEBUG(inode) {}
#define PDEBUG(page, cmd) {}

#endif /* EXT2_OBD_DEBUG */



#define SNAP_ALLOC(ptr, cast, size)					\
do {									\
	if (size <= 4096) {						\
		ptr = (cast)kmalloc((unsigned long) size, GFP_KERNEL);	\
                CDEBUG(D_MALLOC, "kmalloced: %d at %x.\n",		\
		       (int) size, (int) ptr);				\
	} else {							\
		ptr = (cast)vmalloc((unsigned long) size);		\
		CDEBUG(D_MALLOC, "vmalloced: %d at %x.\n",		\
		       (int) size, (int) ptr);				\
	}								\
	if (ptr == 0) {							\
		printk("kernel malloc returns 0 at %s:%d\n",		\
		       __FILE__, __LINE__);				\
	} else {							\
		memset(ptr, 0, size);					\
		snap_kmemory += size;					\
	}								\
} while (0)

#define SNAP_FREE(ptr,size)				\
do {							\
	snap_kmemory -= size;				\
	if (size <= 4096) {				\
		CDEBUG(D_MALLOC, "kfreed: %d at %x.\n",	\
		       (int) size, (int) ptr);		\
		kfree_s((ptr), (size));			\
	} else {					\
		CDEBUG(D_MALLOC, "vfreed: %d at %x.\n",	\
		       (int) size, (int) ptr);		\
		vfree((ptr));				\
	}						\
} while (0)


#ifdef SNAP_DEBUG

extern unsigned int snap_debug_failcode;

#ifdef CONFIG_LOOP_DISCARD
#define BLKDEV_FAIL(dev,fail) loop_discard_io(dev,fail)
#else
#define BLKDEV_FAIL(dev,fail) set_device_ro(dev, 1)
#endif

static inline void snap_debug_device_fail(dev_t dev, unsigned short opcode, unsigned short pos)
{
	unsigned int failcode = (opcode<<16) | pos;

	if( failcode == snap_debug_failcode && !is_read_only(dev)){
		printk(KERN_EMERG "set block device %s into fail mode\n", bdevname(dev));
		BLKDEV_FAIL(dev, 1);
	}
}
#else
#define snap_debug_device_fail(args...) do{}while(0)
#endif

#endif
