#include <stdio.h>
#include <unistd.h>
#include <stdlib.h>
#include <math.h>
#include <time.h>
#include <linux/ext2_fs.h>
#include <ext2fs/ext2fs.h>

ext2_filsys current_fs = NULL;
int total_blocks = 0;
int flags = BLOCK_FLAG_DATA_ONLY;
ino_t root;
ino_t cwd;

/*
 * This routine is used whenever a command needs to turn a string into
 * an inode.
 */
static ino_t string_to_inode(char *str)
{
	ino_t	ino;
	int	len = strlen(str);
	int	i;
	int	retval;

	root = cwd = EXT2_ROOT_INO;

	/*
	 * If the string is of the form <ino>, then treat it as an
	 * inode number.
	 */
	if ((len > 2) && (str[0] == '<') && (str[len-1] == '>')) {
		for (i = 1; i < len-1; i++)
			if (!isdigit(str[i]))
				break;
		if (i == len-1)
			return(atoi(str+1));
	}

	retval = ext2fs_namei(current_fs, root, cwd, str, &ino);
	if (retval) {
//		com_err(str, retval, "");
		return 0;
	}
	return ino;
}

static int list_blocks_proc_line_format(ext2_filsys fs, blk_t *blocknr, int blockcnt, void *private)
{
	printf("%8d     ",blockcnt);	/* Logical block number */
	printf("%8d ", *blocknr);	/* Physical block number */
	printf("\n");
	total_blocks++;
	return 0;
}

static void blockls_dump_blocks(ino_t ino)
{
	printf(" Logical     Physical\n");
	printf("========     ========\n");
	ext2fs_block_iterate(current_fs, ino, flags, NULL,
			list_blocks_proc_line_format, NULL);

	printf("\nTOTAL: %d\n", total_blocks);
}

static inline char* time_to_string(__u32 cl)
{
	time_t t = (time_t) cl;
	return ctime(&t);
}

static void blockls_dump_inode(ino_t inode_num, struct ext2_inode inode)
{
        if (LINUX_S_ISLNK(inode.i_mode) && inode.i_blocks == 0)
                printf("Fast_link_dest: %s\n", (char *)inode.i_block);
        else
                blockls_dump_blocks(inode_num);
}


int blockls_main(int argc, char *argv[])
{
	ino_t ino;
	struct ext2_inode inode_buf;

	if ( argc < 3) {
		printf("usage %s device filename|<ino> flag {all|data(default}\n",argv[0]);
		return 0;
	}	
	if (ext2fs_open(argv[1], EXT2_FLAG_RW, 0, 0, unix_io_manager, &current_fs)){
		printf("can't open file system\n");
		return -1;
	}
	ino = string_to_inode(argv[2]);

	if(argv[3])	{
		if(!strcmp(argv[3], "all"))	flags =0 ;
	}

	if( ext2fs_read_inode(current_fs, ino, &inode_buf) ){
		printf("error reading inode\n");
		return;
	}
	printf("Inode: %d, block list\n", ino);
	blockls_dump_inode(ino, inode_buf);

	ext2fs_close(current_fs);
	return 0;
}
/*-------------------------------------------------------------*/

/*
 * list directory
 */

#define LONG_OPT	0x0001

struct list_dir_struct {
	FILE	*f;
	int	col;
	int	options;
};

static const char *monstr[] = { "Jan", "Feb", "Mar", "Apr", "May", "Jun",
				"Jul", "Aug", "Sep", "Oct", "Nov", "Dec"};
	
static void ls_l_file(struct list_dir_struct *ls, char *name, ino_t ino)
{
	struct ext2_inode	inode;
	errcode_t		retval;
	struct tm		*tm_p;
	time_t			modtime;
	char			datestr[80];

	retval = ext2fs_read_inode(current_fs, ino, &inode);
	if (retval) {
		fprintf(ls->f, "%5ld --- error ---  %s\n", retval, name);
		return;
	}
	modtime = inode.i_mtime;
	tm_p = localtime(&modtime);
	sprintf(datestr, "%2d-%s-%4d %02d:%02d",
		tm_p->tm_mday, monstr[tm_p->tm_mon], 1900 + tm_p->tm_year,
		tm_p->tm_hour, tm_p->tm_min);
 	fprintf(ls->f, "%6ld %6o  %5d  %5d   ", ino, inode.i_mode,
 	       inode.i_uid, inode.i_gid);
 	if (LINUX_S_ISDIR(inode.i_mode))
 		fprintf(ls->f, "%5d", inode.i_size);
 	else
 		fprintf(ls->f, "%5lld", inode.i_size |
			((__u64)inode.i_size_high << 32));
 	fprintf (ls->f, " %s %s\n", datestr, name);
}

static void ls_file(struct list_dir_struct *ls, char *name,
		    ino_t ino, int rec_len)
{
	char	tmp[EXT2_NAME_LEN + 16];
	int	thislen;

	sprintf(tmp, "%ld (%d) %s   ", ino, rec_len, name);
	thislen = strlen(tmp);

	if (ls->col + thislen > 80) {
		fprintf(ls->f, "\n");
		ls->col = 0;
	}
	fprintf(ls->f, "%s", tmp);
	ls->col += thislen;
}	


static int list_dir_proc(struct ext2_dir_entry *dirent,
			 int	offset,
			 int	blocksize,
			 char	*buf,
			 void	*private)
{
	char	name[EXT2_NAME_LEN];

	struct list_dir_struct *ls = (struct list_dir_struct *) private;
	int	thislen;

	thislen = ((dirent->name_len & 0xFF) < EXT2_NAME_LEN) ?
		(dirent->name_len & 0xFF) : EXT2_NAME_LEN;
	strncpy(name, dirent->name, thislen);
	name[thislen] = '\0';

	if (ls->options & LONG_OPT) 
		ls_l_file(ls, name, dirent->inode);
	else
		ls_file(ls, name, dirent->inode, dirent->rec_len);
	
	return 0;
}

void do_list_dir(int argc, char *argv[])
{
	ino_t	inode;
	int	retval;
	struct list_dir_struct ls;
	int	argptr = 1;
	
	ls.f = stdout;
	ls.options = 0;

	if ((argc > argptr) && (argv[argptr][0] == '-')) {
		argptr++;
		ls.options = LONG_OPT;
	}

	if (argc <= argptr)
		inode = cwd;
	else
		inode = string_to_inode(argv[argptr]);
	if (!inode)
		return;

	ls.col = 0;
	retval = ext2fs_dir_iterate(current_fs, inode,
				    DIRENT_FLAG_INCLUDE_EMPTY,
				    0, list_dir_proc, &ls);

	return;
}

static int ls_main(int argc, char *argv[])
{
	if ( argc < 2) {
		printf("usage %s device [dir] \n",argv[0]);
		return 0;
	}	
	if (ext2fs_open(argv[1], EXT2_FLAG_RW, 0, 0, unix_io_manager, &current_fs)){
		printf("can't open file system\n");
		return -1;
	}
	
	root = cwd = EXT2_ROOT_INO;

	do_list_dir(argc-1, ++argv);	
	printf("\n");

	ext2fs_close(current_fs);
	return 0;
}

/*-------------------------------------------------------------*/
int list_block_only = 0;

static int list_blocks_proc(ext2_filsys fs, blk_t *blocknr, int blockcnt, void *private)
{
	printf("%d ", *blocknr);	/* Physical block number */
	total_blocks++;
	return 0;
}

static void stat_dump_blocks(ino_t ino)
{
	if(list_block_only){
		printf(" Logical     Physical\n");
		printf("========     ========\n");
		ext2fs_block_iterate(current_fs, ino, flags, NULL,
				list_blocks_proc_line_format, NULL);
	}else 
		ext2fs_block_iterate(current_fs, ino, flags, NULL,
				list_blocks_proc, NULL);

	printf("\nTOTAL: %d\n", total_blocks);
}

static void stat_dump_inode(ino_t inode_num, struct ext2_inode inode)
{
	const char *i_type;
        char frag, fsize;
	time_t time;
        int os = current_fs->super->s_creator_os;

        if (LINUX_S_ISDIR(inode.i_mode)) i_type = "directory";
        else if (LINUX_S_ISREG(inode.i_mode)) i_type = "regular";
        else if (LINUX_S_ISLNK(inode.i_mode)) i_type = "symlink";
        else if (LINUX_S_ISBLK(inode.i_mode)) i_type = "block special";
        else if (LINUX_S_ISCHR(inode.i_mode)) i_type = "character special";
        else if (LINUX_S_ISFIFO(inode.i_mode)) i_type = "FIFO";
        else if (LINUX_S_ISSOCK(inode.i_mode)) i_type = "socket";
        else i_type = "bad type";
        printf("Inode: %ld   Type: %s    ", inode_num, i_type);
        printf("Mode:  %04o   Flags: 0x%x   Generation: %u\n",
	        inode.i_mode & 0777, inode.i_flags, inode.i_generation);
        printf("User: %5d   Group: %5d   Size: ",
	        inode.i_uid, inode.i_gid);
        if (LINUX_S_ISDIR(inode.i_mode))
		printf("%d\n", inode.i_size);
	else {
		__u64 i_size = (inode.i_size |
				((unsigned long long)inode.i_size_high << 32));
		printf("%lld\n", i_size);
	}
        if (current_fs->super->s_creator_os == EXT2_OS_HURD)
                printf( "File ACL: %d    Directory ACL: %d Translator: %d\n",
		        inode.i_file_acl, LINUX_S_ISDIR(inode.i_mode) ? inode.i_dir_acl : 0,
		        inode.osd1.hurd1.h_i_translator);
        else
                printf( "File ACL: %d    Directory ACL: %d\n",
		        inode.i_file_acl, LINUX_S_ISDIR(inode.i_mode) ? inode.i_dir_acl : 0);
        printf("Links: %d   Blockcount: %d\n", inode.i_links_count,
		                inode.i_blocks);
        switch (os) {
            case EXT2_OS_LINUX:
                frag = inode.osd2.linux2.l_i_frag;
                fsize = inode.osd2.linux2.l_i_fsize;
                break;
            case EXT2_OS_HURD:
                frag = inode.osd2.hurd2.h_i_frag;
                fsize = inode.osd2.hurd2.h_i_fsize;
                break;
            case EXT2_OS_MASIX:
                frag = inode.osd2.masix2.m_i_frag;
                fsize = inode.osd2.masix2.m_i_fsize;
                break;
            default:
                frag = fsize = 0;
        }
        printf( "Fragment:  Address: %d    Number: %d    Size: %d\n",
		inode.i_faddr, frag, fsize);
        printf( "ctime: 0x%08x -- %s", inode.i_ctime,
		time_to_string(inode.i_ctime));
        printf( "atime: 0x%08x -- %s", inode.i_atime,
		time_to_string(inode.i_atime));
        printf( "mtime: 0x%08x -- %s", inode.i_mtime,
		time_to_string(inode.i_mtime));
        if (inode.i_dtime)
		printf("dtime: 0x%08x -- %s", inode.i_dtime,
			time_to_string(inode.i_dtime));
        if (LINUX_S_ISLNK(inode.i_mode) && inode.i_blocks == 0)
                printf("Fast_link_dest: %s\n", (char *)inode.i_block);
        else
                stat_dump_blocks(inode_num);
}


static int stat_main(int argc, char *argv[])
{
	ino_t ino;
	struct ext2_inode inode_buf;

	if ( argc < 3) {
		printf("usage %s device filename|<ino> flag {all|data(default} listblockonly\n",argv[0]);
		return 0;
	}	
	if (ext2fs_open(argv[1], EXT2_FLAG_RW, 0, 0, unix_io_manager, &current_fs)){
		printf("can't open file system\n");
		return -1;
	}
	ino = string_to_inode(argv[2]);

	if(argv[3])	{
		if(!strcmp(argv[3], "all"))	flags =0 ;
	}

	list_block_only = 0;

	if(argv[4]) {
		if(!strcmp(argv[4], "listblockonly"))  list_block_only = 1;
	}
	
	if( ext2fs_read_inode(current_fs, ino, &inode_buf) ){
		printf("error reading inode\n");
		return;
	}
	stat_dump_inode(ino, inode_buf);

	ext2fs_close(current_fs);
	return 0;
}

/*-------------------------------------------------------------*/
int utils_main(int argc, char *argv[])
{
	ino_t ino = 0;
	struct ext2_inode inode_buf;
	FILE *fp;

	if ( argc < 4) {
		printf("usage %s device filename [flag]\n",argv[0]);
		printf("[flag]: -i get inode number from name\n");
		return -1;
	}	
	if (ext2fs_open(argv[1], EXT2_FLAG_RW, 0, 0, unix_io_manager, &current_fs)){
		printf("can't open file system\n");
		return -1;
	}
	if(strcmp(argv[3], "-i") == 0) {
		ino = string_to_inode(argv[2]);
		ext2fs_close(current_fs);

		if(argv[4] != NULL && ino >= 0 ) {
			fp = fopen(argv[4],"w");
			if(fp == NULL) return -1;
			fprintf(fp, "%d", ino);
			fclose(fp);
		}

		return ino;
	}

	ext2fs_close(current_fs);
	return -1;
}

/*-------------------------------------------------------------*/
struct action_struct
{
	const char *act;
	int (*func)(int, char**);
};

struct action_struct actions[] =
{
	{"blockls",	blockls_main},
	{"ls",		ls_main},
	{"stat",	stat_main},
	{"utils",	utils_main}
};
/*-------------------------------------------------------------*/
int main(int argc, char *argv[])
{
	int i;

	if( argc < 2 ){
		printf("usage: %s action [arguments...]\n", argv[0]);
		return -1;
	}

	for( i=0; i<sizeof(actions)/sizeof(struct action_struct); i++ ){
		if( strcmp(argv[1], actions[i].act) == 0 )
			return actions[i].func(argc-1, &argv[1]);
	}

	printf("invalid arguments\n");
	return -1;
}
